{{
*****************************************
* jvmData - JVM memory access functions *
*****************************************
}}


CON

  'Javelin IDE constants
  kReplyBit          = $80
  kFlag              = $7E
  kEscape            = $7D
  kInterrupt         = $19
  kDebugMagic        = $1D
  kRunMagic          = $A4
  kResetWait         = 10000 '// Time taken for SX to reset.
  kAckWait           = 5000  '// The amount of time (in milliseconds) to wait for acknowledgement.
  kResetCharWait     = 3     '// The amount of time to wait before sending first char after a reset.
  kProgramWait       = 2000  '// The amount of time (in milliseconds) to wait for acknowledgement when programming.
  kStepWait          = 20000 '// The amount of time (in milliseconds) to wait for acknowledgement when stepping.
  'kBaseVersion       = $50   '// Compatible base version of the firmware.
  kMaxQueueLength    = 500   '// Maximum number of characters than can be queued.
  kComNone           = 0
  kComStatus         = 1
  kComReset          = 2
  kComStep           = 3
  kComRun            = 4
  kComError          = 5
  kComStop           = 6
  kComStack          = 7
  kComConfig         = 8
  kComHeap           = 9
  kComBreakpoints    = 10
  kComProgram        = 11
  kComDownloadOK     = 12
  kComStepDone       = 13
  kComAtn            = 14
  kComStartLoad      = 15
  'kComEndLoad        = kBaseVersion
  kComTerminalQuery  = 17     '// Check for data from the terminal.
  kComTerminalData   = 18     '// Data follows.
  kComTerminalNoData = 19     '// No data follows.
  kComDownloadRamOK  = 20

  'activity values
  kPassive = 0   'JVM was running but has stopped (exception,end)
  kRun     = 1   'JVM running
  kJog     = 2   '???
  kBreak   = 3   'JVM reached breakpoint
  kEnd     = 4   'java program ended
  
  'values following kComError
  kErrNone       = 0
  kErrOpcode     = 1
  kErrBreakpoint = 2
  kErrMessage    = 3   'to send message to IDE message window
  kErrException  = 4   'to send exception to IDE status window
  kErrEnd        = 5   'to signal end of java program

  'exceptions
  kIllegalArgumentException  = 1
  kNullPointerException      = 2
  kIndexOutOfBoundsException = 3
  kOutOfMemoryError          = 4

CON

  '// The SX banks the VPs use.
  nmPeriph0Bank = $80
  nmPeriph1Bank = nmPeriph0Bank + $10
  nmPeriph2Bank = nmPeriph0Bank + $20
  nmPeriph3Bank = nmPeriph0Bank + $30
  nmPeriph4Bank = nmPeriph0Bank + $40
  nmPeriph5Bank = nmPeriph0Bank + $50
  nmBlockBank   = $70

  nmPortA       = $06  'corresponds to I/O pins 0..7
  nmPortB       = $07  'corresponds to I/O pins 8..15
  'reserved values for extending Javelin code to PropJava code
  nmPortC       = $08  'corresponds to I/O pins 16..23
  nmPortD       = $09  'corresponds to I/O pins 24..31
  
  pin0  = (1<<0)         'A constant representing pin P0
  pin1  = (1<<1)         'A constant representing pin P1
  pin2  = (1<<2)         'A constant representing pin P2
  pin3  = (1<<3)         'A constant representing pin P3 
  pin4  = (1<<4)         'A constant representing pin P4 
  pin5  = (1<<5)         'A constant representing pin P5
  pin6  = (1<<6)         'A constant representing pin P6 
  pin7  = (1<<7)         'A constant representing pin P7 
  pin8  = (1<<0) + $100  'A constant representing pin P8
  pin9  = (1<<1) + $100  'A constant representing pin P9
  pin10 = (1<<2) + $100  'A constant representing pin P10
  pin11 = (1<<3) + $100  'A constant representing pin P11 
  pin12 = (1<<4) + $100  'A constant representing pin P12 
  pin13 = (1<<5) + $100  'A constant representing pin P13
  pin14 = (1<<6) + $100  'A constant representing pin P14
  pin15 = (1<<7) + $100  'A constant representing pin P15  
  'reserved values for extending Javelin code to PropJava code
  pin16 = (1<<0) + $200  'A constant representing pin P16
  pin17 = (1<<1) + $200  'A constant representing pin P17
  pin18 = (1<<2) + $200  'A constant representing pin P18
  pin19 = (1<<3) + $200  'A constant representing pin P19 
  pin20 = (1<<4) + $200  'A constant representing pin P20 
  pin21 = (1<<5) + $200  'A constant representing pin P21
  pin22 = (1<<6) + $200  'A constant representing pin P22
  pin23 = (1<<7) + $200  'A constant representing pin P23  
  pin24 = (1<<0) + $300  'A constant representing pin P24
  pin25 = (1<<1) + $300  'A constant representing pin P25
  pin26 = (1<<2) + $300  'A constant representing pin P26
  pin27 = (1<<3) + $300  'A constant representing pin P27 
  pin28 = (1<<4) + $300  'A constant representing pin P28 
  pin29 = (1<<5) + $300  'A constant representing pin P29
  pin30 = (1<<6) + $300  'A constant representing pin P30
  pin31 = (1<<7) + $300  'A constant representing pin P31  

  'native registers

  'file stamp\core\ADC.java (offsets within a VP bank) 
  nmAdcValue = $0B  '// Current ADC value.
  nmInPort   = $0C  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
  nmInPin    = $0D  '// bitmask (1 of 8 set)
  nmOutPort  = $0E  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
  nmOutPin   = $0F  '// bitmask (1 of 8 set)

  'file stamp\core\DAC.java (offsets within a VP bank) 
  nmDacValue = $0B  '// Current ADC value.
  nmDacPort  = $0E  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
  nmDacPin   = $0F  '// bitmask (1 of 8 set)

  'shiftout modes
  SHIFT_MSB = 0  'Specify MSB-first transmission.
  SHIFT_LSB = 2  'Specify LSB-first transmission. 

  'shiftin modes
  PRE_CLOCK_MSB  = 0  'Specify pre-clock sampling and MSB-first transmission. 
  POST_CLOCK_MSB = 1  'Specify post-clock sampling and MSB-first transmission.
  PRE_CLOCK_LSB  = 2  'Specify pre-clock sampling and LSB-first transmission.
  POST_CLOCK_LSB = 3  'Specify post-clock sampling and LSB-first transmission.
  OnClock        = 4

  'file stamp\core\EEPROM.java
  JVM_SB_MSB = $4C  '// Highbyte of stackbase (first free eeprom address)
  JVM_SB_LSB = $4D  '// Lowbyte of stackbase (first free eeprom address)
  {{
     The EEPROM.java methods size(),write() and read() use $7FFF for eeprom size.
     On the prop the eeprom size is set to JAVASIZE. To let the methods access the
     correct addresses, we must adjust for this difference. That means the value
     from register JVM_SB must be increased by JVM_SB_ADJUST. Originally JVM_SB is
     set to the stack base as found on eeprom address $0012.
  }}

  'file stamp\core\Memory.java
  JVM_HEAP_POINTER_LSB  = $18   '//jvm heap pointer (word)
  JVM_HEAP_POINTER_MSB  = $19
  JVM_STACK_POINTER_LSB = $12   '//jvm stack pointer (word)
  JVM_STACK_POINTER_MSB = $13

  {{
         Stackframe format
         +---------------+
         | Local0        | <- JVM_FP
         | ...           |
         | LocalN        |
         +---------------+
         | Method MSB    | <- JVM_FP + numLocals
         | Method LSB    |
         +---------------+
         | Old FP MSB    |
         | Old FP LSB    |
         +---------------+
         | JVM_PC MSB    |
         | JVM_PC LSB    |
         +---------------+
         | Stack0        |
         | ....          |
         +---------------+
                           <- JVM_SP
  }}

  kMPOffset = 0   '//method pointer offset
  kFPOffset = 2   '//frame pointer offset
  kPCOffset = 4   '//program counter offset

  {{
         Memory map layout
  $0000  +---------------+ @javaProg
         |    header     |
  $0014  +---------------+
         | jemcodes      |
         | ...           |
  $XXX0  +---------------+ stackbase                 ---
         | Local0        | <- JVM_FP                  ^
         | ...           |                            |
         | LocalN        |                            |
         +---------------+                            |
         | Method MSB    | <- JVM_FP + numLocals      |
         | Method LSB    |                            |
         +---------------+                            |
         | Old FP MSB    |                            |
         | Old FP LSB    |
         +---------------+                       stack frame
         | JVM_PC MSB    |                            
         | JVM_PC LSB    |                            | 
         +---------------+                            |
         | Stack0        | operand stack              |
         | ....          |                            V
         +---------------+                           ---
         |               | <- JVM_SP
         |     free      |
         |               |

         |               |
         |     free      |
         |               |
         +---------------+
         |               | <- JVM_HP
         |   allocated   |
         |               |
         +---------------+
         |  static vars  | size := byte[@javaProg + JVM_STATIC]   'size in bytes
         |  each var     | if size == 0 
         |    2 bytes    |   size := 256
         +---------------+                           
  JAVASIZE                 heapbase (on javelin $7FFF)

  Header layout:
  
  0000:   dw      $A4     ; Placeholder for run/debug.
  0001:   dw      $0      ; Placeholder for the length.
  0002:   dw      $01     ; Version: 0.1
  0003:   dw      16      ; Number of static fields = 8 (maximum of 128 static fields)
  0004:   dw      $00     ; `main' method: Offset hi = $00
  0005:   dw      $2C     ;                       lo = $2C
  0006:   dw      $00     ; java.lang.Object class: Offset hi = $00
  0007:   dw      $8D     ;                                lo = $8D
  0008:   dw      $01     ; java.lang.String class: Offset hi = $01
  0009:   dw      $59     ;                                lo = $59
  000A:   dw      $00     ; Throwable.throwVMException method: Offset hi = $00
  000B:   dw      $52     ;                                           lo = $52
  000C:   dw      $09     ; String table offset hi = $09
  000D:   dw      $7F     ;                     lo = $7F
  000E:   dw      $00     ; Callback method offset hi = $00
  000F:   dw      $00     ;                        lo = $00
  0010:   dw      $01     ; java.lang.OutOfMemoryError class: Offset hi = $01
  0011:   dw      $2C     ;                                          lo = $2C
  0012:   dw      $09     ; Stack base hi = $09
  0013:   dw      $B0     ;            lo = $B0
  
  }}

  'javaProg header fields
  JVM_RUNDEBUG     = $0000
  JVM_LENGTH       = $0001
  JVM_VERSION      = $0002
  JVM_STATIC       = $0003
  JVM_MAIN         = $0004
  JVM_OBJECT       = $0006
  JVM_STRING       = $0008
  JVM_THROW        = $000A
  JVM_STRTABLE     = $000C
  JVM_CALLBACK     = $000E
  JVM_MEMERROR     = $0010
  JVM_STKBASE      = $0012

  'class header fields
  JVM_SUPER        = $00
  JVM_FIELDS       = $02
  
  'method header fields
  JVM_PARAMETERS    = $00
  JVM_LOCALS        = $01
  JVM_FLAGS         = $02
  JVM_CODE          = $03
  JVM_EXCEPTION     = $05
  

CON
  BREAKPOINTS = 10     'number of breakpoints supported by debugger
  MAX_NUM_VPS = 6      'Number of available VP slots

   
DAT
  javaProg       word 0                   'address of javaProg storage area
  javaSize       word 0                   'size of javaProg
  
  jvmHP          word 0                   'heap pointer
  jvmSP          word 0                   'stack pointer
  jvmSB          word 0                   'stack base (also first free eeprom address)
  jvmFP          word 0                   'frame pointer
  jvmPC          word 0                   'program counter
  jvmMP          word 0                   'method pointer
  jvmEP          word 0                   'exception pc
  
  memSize        word 0                   'requested by QueryConfig
  progSize       word 0                   'requested by QueryConfig
  heapBase       word 0                   'requested by QueryConfig
  staticBase     word 0                   'base of static variables area
  stringIndex    word 0                   'address of string index table
  eeBase         word 0                   'requested by stamp\core\EEPROM.java
                                          '  set to return true available free eeprom space
                                       
  bpList         word 0[BREAKPOINTS+1]    'list of breakpoints (entries are jvmPC address values)
  numBreakpoints byte 0                   'number of breakpoints
  activity       byte 0                   'requested by QueryStatus
  carry          byte 0                   'set to true if math expression overflows
  kComEndLoad    byte 0                   'holds device version ($50,$70,$78)
  notifyStepDone byte 0                   'set to true if jvm must notify IDE after step is done 
  mustDoStep     byte 0                   'set to true when jvm must do step

  
  
PUB writeJVMregister(address,value)
{{ write value to register }}
  address &= $FF
  case address
    nmPortA: outa[7..0] := value
    nmPortB: outa[15..8] := value
    nmPortC: outa[23..16] := value
    nmPortD: outa[31..24] := value

  
PUB readJVMregister(address):value
{{ read value from register }}
  value := 0
  address &= $FF
  case address
    nmPortA: value := ina
    nmPortB: value := ina >> 8
    nmPortC: value := ina >> 16
    nmPortD: value := ina >> 24
    JVM_SB_MSB: value := (($8000 - javaSize) + jvmSB) >> 8  '// Highbyte of simulated stackbase (first free eeprom address)
    JVM_SB_LSB: value := (($8000 - javaSize) + jvmSB)       '// Lowbyte of simulated stackbase (first free eeprom address)
    JVM_HEAP_POINTER_MSB: value := jvmHP >> 8  '//jvm heap pointer (highbyte)
    JVM_HEAP_POINTER_LSB: value := jvmHP       '//jvm heap pointer (lowbyte)
    JVM_STACK_POINTER_MSB: value := jvmSP >> 8 '//jvm stack pointer (highbyte)
    JVM_STACK_POINTER_LSB: value := jvmSP      '//jvm stack pointer (lowbyte)
  value &= $FF

  